package teikade.debug;

import java.net.*;
import java.io.*;
import java.util.*;
import teikade.sys.SystemManager;
import teikade.sys.TeikadeProperties;
import teikade.net.*;
import teikade.util.Terminal;

public class DebuggerManager extends Observable implements Runnable {
    ServerSocket server;
    Socket socket;
    InputStream fromDebugger;
    PrintStream toDebugger;
    Terminal term = new Terminal();
    Vector breakPoints = new Vector();
    String runClass = null;
    private String classNamePrefix = "teikade.debug.";
    private static Vector debuggers = new Vector();
    private static final int startFrom = 7700;
    private static final int before = 7800;
    private static int port = startFrom;
    private int portNumber;

    public void clearBreakPoints() {
	Packet packet = new ClearBreakPoints();
	sendPacket(packet);
	clearNotify(packet);
    }

    public void start() throws IOException {
	if (server == null) {
	    server = new ServerSocket(portNumber);
	    debuggers.addElement(this);
	    new Thread(this,this.getClass().getName()).start();
	}
    }

    static private int getPortNumber() {
	if (port < before)
	    return port++;
	else
	    return -1;
    }

    public static Enumeration list() {
	return debuggers.elements();
    }

    public DebuggerManager(int port) {
	portNumber = port;
    }

    public static DebuggerManager createManager() throws IOException {
	return createManager(TeikadeProperties.getParsedString("command.debugger"));
    }

    public static DebuggerManager createManager(String command) throws IOException {
	int no;
	while ((no = getPortNumber()) != -1) {
	    try {
		DebuggerManager manager = new DebuggerManager(no);
		manager.setCommand(command + no);
		manager.start();
		return manager;
	    } catch (IOException e) {}
	}
	throw new IOException("port number exhausted");
    }

    public void setCommand(String command) {
	term.setCommand(command);
    }

    public static void setDefaultCommand(String command) {
	// not implemented yet
	// TeikadeProperties.setProperty("command.debugger", command);
    }

    public void run() {
	term.show();
	try {
	    term.start();
	} catch (IOException e) {
	    SystemManager.information(e.getMessage());
	    return;
	}
	while (server != null) {
	    try {
		socket = server.accept();
		fromDebugger = new BufferedInputStream(socket.getInputStream());
		toDebugger = new PrintStream(
			new BufferedOutputStream(socket.getOutputStream()));
	    } catch (IOException e) {
		SystemManager.information("Couldn't set up connection to the debugger");
		continue;
	    }
	    while (socket != null) {
		try {
	            Packet packet = Packet.from(fromDebugger, classNamePrefix);
		    if (packet == null) {
			break;
		    } else if (packet instanceof DebuggerManagerPacket) {
			packet.execute(this);
		    } else {
			setChanged();
			notifyObservers(packet);
		    }
		} catch (IllegalArgumentException e) {
		    // SystemManager.information(e.getMessage());
		    e.printStackTrace();
		}
	    }
	    closeConnection();
	    if (!term.isVisible()) {
		server = null;
	    }
	}
	debuggers.removeElement(this);
    }

    public String name() {
	return server.toString();
    }

    public String getClassNamePrefix() {
	return classNamePrefix;
    }

    public void setClassNamePrefix(String prefix) {
	classNamePrefix = prefix;
    }

    public void setBreakPoint(String className, int lineNo) {
	Packet packet = new SetBreakPoint(className, lineNo);
	sendPacket(packet);
	String args = packet.argumentsToString();
	if (!breakPoints.contains(args)) {
	    breakPoints.addElement(args);
	}
    }

    public void resetBreakPoint(String className, int lineNo) {
	Packet packet = new ResetBreakPoint(className, lineNo);
	sendPacket(packet);
	String args = packet.argumentsToString();
	if (breakPoints.contains(args)) {
	    breakPoints.removeElement(args);
	}
    }

    public void breakPointNotify(Packet packet, boolean set) {
	String args = packet.argumentsToString();
	boolean contains = breakPoints.contains(args);
	if (set && !contains) {
	    breakPoints.addElement(args);
	} else if (!set && contains) {
	    breakPoints.removeElement(args);
	}
    }

    public void clearNotify(Packet packet) {
	breakPoints.removeAllElements();
    }

    public void runClass(String className, String classPath, String args) {
	Packet packet = new RunClass(className, classPath, args);
	sendPacket(packet);
    }

    public void runApplet(String className, String classPath, String html) {
	Packet packet = new RunApplet(className, classPath, html);
	sendPacket(packet);
    }

    void listBreakPoints() {
	Enumeration e = breakPoints.elements();
	while (e.hasMoreElements()) {
	    String args = (String)e.nextElement();
	    sendPacket("SetBreakPoint(" + args + ")");
	}
    }

    void closeConnection() {
	if (socket != null) try {
	    socket.close();
	} catch (IOException e) {
	    // should not complain about closing connection...
	}
	socket = null;
    }

    private void sendPacket(String packetString) {
	try {
	    if (!isConnected()) start();
	    toDebugger.println(packetString);
	    toDebugger.flush();
	} catch (IOException e) {
	}
    }

    private void sendPacket(Packet packet) {
	sendPacket(packet.toString());
    }

    public boolean isConnected() {
	return socket != null;
    }

    /**
     * returns local port on which the debugger manager is listening
     */
    public int getLocalPort() {
	return (server != null) ? server.getLocalPort() : -1;
    }
}
