/*
 * Copyright (c) 1995-1998 PFU Limited.
 *      author Osamu Satoh, Kazuhisa Shirakami
 */
package teikade.gui;

import teikade.sys.*;
import teikade.util.*;
import teikade.gui.component.*;
import teikade.gui.menu.*;
import java.awt.*;
import java.util.*;
import java.io.File;
import java.io.DataInputStream;
import java.io.IOException;

public class CompileFrame extends CommandFrame implements CompileListener {
    private TextArea errorMessage;
    private boolean compileUpdated;
    private boolean noError;
    private String compileList;
    private static final boolean AUTO_CLOSE = "true".equals(
	    TeikadeProperties.getString("attribute.compileframe.autoclose"));

    private static final Hashtable frames = new Hashtable();

    static void openOn(CodePath cp, boolean updated) {
        CompileFrame compile = find(cp);
        if (compile == null) {
            compile = new CompileFrame(cp, "Compile Frame on: " + cp.printString(), updated);
            compile.start();
            compile.open();
        } else {
	    compile.setUpdated(updated);
            compile.show();
            compile.restart();
	    compile.requestFocus();
        }
    }

    public void setUpdated(boolean updated) {
	compileUpdated = updated;
    }

    public CompileFrame(CodePath cp, String title, boolean updated) {
	super(cp, title, "compilation");
	compileUpdated = updated;
	frames.put(cp.printString(), this);
    }

    protected TeikadeMenuBar createMenuBar() {
	return new CompileMenuBar();
    }

    protected Component createListingItems() {
	errorMessage = new DTextArea(4, 80);
	errorMessage.setEditable(false);
	Panel panel = new Panel();
	panel.setLayout(new BorderLayout());
	panel.add(CENTER, super.createListingItems());
	panel.add(SOUTH, errorMessage);
	return panel;
    }

    public void start() {
	start(Compile.getInstance(startup, compileUpdated));
    }

    public void start(Command cmd) {
	Compile compile = (Compile)cmd;
	noError = true;
	if (!acceptWithConfirm()) {
	    information("Compilation on " + startup.printString() + " aborted.");
	    return;
	}
	compile.setCompileListener(this);
	CompileTarget target = compile.getCompileTarget();
	int num = target.getItemsCount();
	if (num == 0) {
	    addMessage("No source files to be compiled.\n");
	    return;
	} else if (num == 1) {
	    compileList = "Compile file: " + target.getItemAt(0);
	} else {
	    StringBuffer message = new StringBuffer();
	    message.append("Compile files: ");
	    for (int i = 0; i < num; i++) {
		message.append('\n').append(target.getItemAt(i));
	    }
	    compileList = message.toString();
	}
	super.start(compile);
    }

    public void restart() {
	noError = true;
	restart(Compile.getInstance(startup, compileUpdated));
    }

    public void stop() {
	noError = false;
	super.stop();
    }

    protected boolean select(int newSelection) {
	try {
	    if (super.select(newSelection)) {
		if (newSelection == -1) {
		    errorMessage.setText(compileList);
		} else {
		    errorMessage.setText(((CompileCodeLocation)itemAt(newSelection)).detailMessage());
		}
		return true;
	    }
	    return false;
	} catch (TeikadeException e) {
	    information(e.getMessage());
	    return false;
	}
    }

    protected void update() {
	super.update();
	((CompileMenuBar)menuBar).updateExecuteMenuActivity(startup);
    }

    public CodePath getExecuteCodePath() {
	return startup;
    }

    protected ListingFrame copyCreate(CodePath cp) {
	return new CompileFrame(cp, "Same elements for: " + cp.printString(), compileUpdated);
    }

    protected static CompileFrame find(CodePath cp) {
	if (frames.containsKey(cp.printString())) {
	    return (CompileFrame)frames.get(cp.printString());
	} else {
	    return null;
	}
    }

    /**
     * Thanks for auto close patch by Chris Normand 1997.05.07
     *
     * Override run().  After the run is completed, check the error flag.
     * If there are no errors, close the window.
     * Delay a bit so the user can be sure there were no errors (for their
     * peace of mind).
     */
    public void run() {
	super.run();
	if (AUTO_CLOSE && noError) {
	    try {
		Thread.sleep(150);
	    } catch (InterruptedException e) {
	    }
	    close();
	}
    }

    public boolean close() {
	if (!super.close()) return false;
	frames.remove(startup.printString());
	return true;
    }

    /**
     * teikade.util.CompileListener stuff
     */
    public void result(Compile compile, CompileResult result)  {
	try {
	    if (result.type() == CompileResult.MESSAGE) {
		addMessage(result.message());
	    } else {
		StringBuffer selection = new StringBuffer();
		String fname = result.fileName();
		if (fname != null) {
		    int sep = fname.lastIndexOf(File.separatorChar);
		    if (sep > 0) {
			if (fname.toLowerCase().endsWith(".java")) {
			    selection.append(fname.substring(sep + 1, fname.length() - 5));
			} else {
			    selection.append(fname.substring(sep + 1));
			}
		    } else {
			selection.append(fname);
		    }
		    selection.append(':');
		    selection.append(result.line());
		    selection.append(": ");
		}
		selection.append(result.description());
		CompileCodeLocation item = new CompileCodeLocation(
		    result.fileName(),
		    result.line(),
		    selection.toString(),
		    result.offset(),
		    0,
		    result.message());
		addItem(item);
		list.reshape();
	    }
	} catch (CodeLocationException e) {
	    addMessage("fail to search location " +
		       result.fileName() + ":" + result.line() + ", sorry.\n");
	}
    }
}

class CompileCodeLocation extends CodeLocation {
    private String detailMessage;

    public CompileCodeLocation(String fname, int line, String item, int offset, int length, String msg) 
                                                       throws CodeLocationException {
	super(fname, line, item, offset, length);
	detailMessage = msg;
    }

    public CompileCodeLocation(CodeLocation clone) {
	super(clone);
	detailMessage = ((CompileCodeLocation)clone).detailMessage();
    }

    public String detailMessage() {
	return detailMessage;
    }

    public ListItem cloneItem() {
	return new CompileCodeLocation(this);
    }
}
