/*
 * Copyright (c) 1995 PFU Limited.
 *	author Osamu Satoh, Kazuhisa Shirakami
 */
package teikade.gui;

import teikade.sys.*;
import teikade.util.*;
import teikade.gui.menu.*;
import teikade.gui.component.*;
import teikade.gui.editor.EditData;
import teikade.gui.component.SplitPanel;

import java.awt.*;
import java.util.Date;
import java.util.Vector;
import java.util.Enumeration;

public class ListingFrame extends TeikadeFrame {
    protected DList list;
    private Vector items = new Vector();
    protected int selection = -1;
    private StringBuffer messages = new StringBuffer();
    private ListingReplacer listingReplacer = null;

    ListingFrame(CodePath cp, String title) {
	super(cp, title);
    }

    protected TeikadeMenuBar createMenuBar() {
	return new ListingMenuBar();
    }

    protected Component createDisplayItems() {
	SplitPanel panel = new SplitPanel(SplitPanel.VERTICAL);
	panel.add(createListingItems());
	panel.addSplitHandle();
	panel.add(super.createDisplayItems());
	return panel;
    }

    protected Component createListingItems() {
	list = new DList(TeikadeProperties.getInteger("attribute.listing.lines"), false);
	return list.makePanel();
    }

    /**
     * return current CodePath
     */
    public CodePath codePath() {
	return new CodePath(current);
    }

    public int getSelection() {
	return selection;
    }

    public boolean handleEvent(Event ev) {
	if (ev.target == list) {
	    int selected = list.getSelectedIndex();
	    switch (ev.id) {
	    case Event.LIST_SELECT:
	    case Event.LIST_DESELECT:
		try {
		    select(selected);
		} catch (TeikadeException e) {
		    information(e.getMessage());
		}
		if (SystemManager.onWin32) codeData.requestFocus();
		return true;
	    case Event.ACTION_EVENT:
		if (isMarked(selected)) {
		    unmark(selected);
		} else {
		    mark(selected);
		}
		if (selection != selected) {
		    try {
			select(selected);
		    } catch (TeikadeException e) {
			information(e.getMessage());
		    }
		}
		return true;
	    }
	}
	return super.handleEvent(ev);
    }

    protected boolean select(int newSelection)  throws TeikadeException {
	CodePath cp;
	if (newSelection < 0 || newSelection == selection) {
	    cp = CodePath.nullCodePath;
	    newSelection = -1;
	} else {
	    ListItem item = itemAt(newSelection);
	    cp = item.codePath();
	    if (selection != -1 && item.isSameCode(itemAt(selection))) {
		// same code, no need to set code, move cursor only
	        selectCommonAction1( newSelection, item );
		return true;
	    }
	}
	if (acceptWithConfirm()) {
	    selectCommonAction2( newSelection, cp );
	    return true;
	} else {
	    selectCommonAction3();
	    return false;
	}
    }

    void selectCommonAction1(int newSelection, ListItem item) {
        selection = newSelection;
        list.selectAndVisible(selection);
        int cursor[] = item.cursorPos();
        codeData.select(cursor[0], cursor[1]);
    }

    void selectCommonAction2(int newSelection, CodePath cp) {
        selection = newSelection;
        current = cp;
        list.selectAndVisible(selection);
        updateCurrent();        
    }

    void selectCommonAction3() {
        if (selection < 0) {
	    list.deselect(list.getSelectedIndex());
        } else {
	    list.select(selection);
        }
    }

    /**
     * update
     */
    protected void update() {
	if (selection < 0) {
	    int selected = list.getSelectedIndex();
	    if (selected > 0) list.deselect(selected);
	} else {
	    if (!list.isSelected(selection)) list.selectAndVisible(selection);
	}
	super.update();
    }

    protected void updateCode() {
        if (selection == -1) {
            setCode(messages.toString());
	} else {
	    DList save = list;
	    list = null;
	    super.updateCode();
	    list = save;
	    ListItem item = (ListItem)itemAt(selection);
	    int cursor[] = item.cursorPos();
	    codeData.select(cursor[0], cursor[1]);
	}
    }

    protected void updateButtonActivity() {
	if (selection == -1) {
	    updateButtonActivity(false, false);
	} else {
	    ListItem item = (ListItem)itemAt(selection);
	    updateButtonActivity(item.isChanged() || codeData.isEditable(),
				codeData.isEditable());
	}
    }

    /**
     * set mark to specified ListItem
     */
    public void mark(int pos) {
	ListItem item = itemAt(pos);
	if (!item.isMarked()) {
	    item.setMark(true);
	    updateList();
	}
	update();
    }

    /**
     * reset mark to specified ListItem
     */
    public void unmark(int pos) {
	ListItem item = itemAt(pos);
	if (item.isMarked()) {
	    item.setMark(false);
	    updateList();
	}
	update();
    }

    /**
     * return marked or not of specified ListItem
     */
    public boolean isMarked(int pos) {
	return itemAt(pos).isMarked();
    }

    /**
     * set mark to all ListItem
     */
    public void markAll() {
	for (int pos = 0; pos < items.size(); pos++) {
	    ListItem item = itemAt(pos);
	    if (!item.isMarked()) {
		item.setMark(true);
	    }
	}
	updateList();
	update();
    }

    /**
     * reset mark to all ListItem
     */
    public void unmarkAll() {
	for (int pos = 0; pos < items.size(); pos++) {
	    ListItem item = itemAt(pos);
	    if (item.isMarked()) {
		item.setMark(false);
	    }
	}
	updateList();
	update();
    }

    void updateList() {
	list.clear();
	for (int pos = 0; pos < items.size(); pos++) {
	    ListItem item = itemAt(pos);
	    list.addItem((item.isMarked() ? "M " : "  ") + item.listString());
	}
	list.reshape();
	if (selection != -1) {
	    list.selectAndVisible(selection);
	}
    }

    protected Enumeration itemElements() {
	return items.elements();
    }

    protected boolean addItem(ListItem item) {
	items.addElement(item);
	list.addItem((item.isMarked() ? "M " : "  ") + item.listString());
	return true;
    }

    protected ListItem itemAt(int pos) {
	return (ListItem)items.elementAt(pos);
    }

    public void spawnMatchingItems(CodePath cp, boolean exact) {
	ListingFrame newFrame = copyCreate(cp);
	boolean open = false;
	for (int pos = 0; pos < items.size(); pos++) {
	    ListItem item = itemAt(pos);
	    if (item.codePath().matches(cp, exact)) {
		item = item.cloneItem();
		if (item != null) {
		    newFrame.addItem(item);
		    open = true;
		}
	    }
	}
	if (open) newFrame.open();
    }

    public void markMatchingItems(CodePath cp, boolean exact, boolean mark) {
	int pos;
	for (pos = 0; pos < items.size(); pos++) {
	    ListItem item = itemAt(pos);
	    if (item.codePath().matches(cp, exact)) {
		item.setMark(mark);
	    }
	}
	updateList();
    }

    public void spawnMarkedItems(boolean marked) {
	ListingFrame newFrame = copyCreate(startup);
	boolean open = false;
	for (int pos = 0; pos < items.size(); pos++) {
	    ListItem item = itemAt(pos);
	    if (item.isMarked() == marked) {
		item = item.cloneItem();
	        if (item != null) {
		    newFrame.addItem(item);
		    open = true;
		}
	    }
	}
	if (open) newFrame.open();
    }

    public void start(Enumeration items) {
	while (items.hasMoreElements()) {
	    addItem((ListItem)items.nextElement());
	}
	list.reshape();
    }

    protected boolean syncAndClearListItems() {
	try {
	    if (!select(-1)) {
		SystemManager.information("Could not restart compilation:\n" + startup.printString());
		return false;
	    }
	} catch (TeikadeException e) {
	    information(e.getMessage());
	    return false;
	}
	startup.getManager().sync();
	CodePath syncedCp = startup;
	Enumeration cpList = items.elements();
	while (cpList.hasMoreElements()) {
	    CodePath cp = ((ListItem)cpList.nextElement()).codePath();
	    if (cp != null && !cp.matches(syncedCp)) {
		cp.getManager().sync();
		syncedCp = cp;
	    }
        }
	list.clear();
	items.setSize(0);
	messages.setLength(0);
	update();
	return true;
    }

    protected void restart(Enumeration newItems) {
	if (syncAndClearListItems()) {
	    start(newItems);
	}
    }

    protected String getMessages() {
	return messages.toString();
    }

    protected void addMessage(String message) {
	if (message == null || message.length() <= 0) return;
	messages.append(message);
	if (selection == -1) {
            setCode(messages.toString());
	}
    }

    protected void addMessageWithDate(String message) {
	addMessage("--- " + message + ": " + (new Date()).toString() + "\n");
    }

    public boolean close() {
	if (!super.close()) return false;
	if (listingReplacer != null) {
	    listingReplacer.hide();
	    listingReplacer.dispose();
	    listingReplacer = null;
	}
	return true;
    }

    protected ListingFrame copyCreate(CodePath cp) {
	return new ListingFrame(cp, "Same elements for: " + cp.printString());
    }

    public void replace(EditData editData) {
	replace(editData, null);
    }

    public void replace(EditData editData, String searchString) {
	if (listingReplacer == null) {
	    listingReplacer = new ListingReplacer(this, editData);
	}
	if (searchString != null) {
	    listingReplacer.setSearchString(searchString);
	}
	listingReplacer.show();
    }
}
