/*
 * Copyright (c) 1995 PFU Limited.
 *	author Kazuhisa Shirakami
 */
package teikade.gui.dialog;

import teikade.sys.*;
import teikade.util.StringSorter;
import teikade.gui.component.*;
import java.io.File;
import java.io.IOException;
import java.awt.*;

public class TeikadeFileDialog extends TeikadeDialog {
    private TextField pathField;
    private Button chooseButton;
    private TextField postfixField;
    private Button postfixButton;
    private Component chooser;
    private DList dirs;
    private DList files;
    private int lastDirSelected = -1;
    private int lastFileSelected = -1;

    private static final String CURRENT = "." + File.separator;
    private static final String PARENT = ".." + File.separator;
    private String currentDir = SystemManager.USERHOME;
    private String currentFile = "";
    private String startFile   = "";

    private String chosenFile;

    public TeikadeFileDialog(Frame f, String title) {
	this(f, title, true);
    }

    public TeikadeFileDialog(Frame f, String title, boolean withDirectory) {
	super(f, title);
	chooser = createChooser(withDirectory);
	chooser.hide();
	Panel panel = new Panel();
	panel.setLayout(new BorderLayout());
	panel.add(NORTH, createDisplayItems());
	panel.add(CENTER, chooser);
	add(CENTER, panel);
	setDirectory(null);
    }

    protected Component createDisplayItems() {
	Panel panel = new Panel();
	panel.setLayout(new BorderLayout());
	panel.add(CENTER, pathField = new DTextField(40));
	panel.add(EAST, chooseButton = new DButton("choose"));
	return panel;
    }

    protected Component createChooser(boolean withDirectory) {
	Panel chooser = new Panel();
	chooser.setLayout(new BorderLayout());

	Panel postfixArea = new Panel();
	postfixArea.setLayout(new BorderLayout());
	postfixArea.add(WEST, new DLabel("Postfix:"));
	postfixArea.add(CENTER, postfixField = new DTextField());
	postfixArea.add(EAST, postfixButton = new DButton("set postfix"));
	postfixField.setEditable(true);

	int listLines = TeikadeProperties.getInteger("attribute.files.lines");
	if (withDirectory) {
	    dirs = new DList(listLines, false);
	    Panel dirsArea = new Panel();
	    dirsArea.setLayout(new BorderLayout());
	    dirsArea.add(NORTH, new DLabel("Directories:"));
	    dirsArea.add(CENTER, dirs.makePanel());
	    chooser.add(WEST, dirsArea);
	}
	files = new DList(listLines, false);
	Panel filesArea = new Panel();
	filesArea.setLayout(new BorderLayout());
	filesArea.add(NORTH, new DLabel("Files:"));
	filesArea.add(CENTER, files.makePanel());

	chooser.add(CENTER, filesArea);
	chooser.add(SOUTH, postfixArea);
	return chooser;
    }

    protected String getPath() {
	String path = pathField.getText();
	int top = 1;	// for root directory
	if (SystemManager.onWin32 || SystemManager.onOS2) {
	    top = path.indexOf(':') + 2;
	    if (top < 2) top = 1;
	}
	if (path.length() > top) {
	    while (path.endsWith(File.separator)) {
		path = path.substring(0, path.lastIndexOf(File.separator));
	    }
	}
	return path;
    }

    public void setDirectory(String dir) {
	select(dir);
    }

    public void setFile(String file) {
	startFile = file;
	currentFile = "";
	select(currentDir + file);
    }

    public void setPostFix(String postFix) {
	postfixField.setText(postFix);
	select(null);
    }

    public String request() {
	if (!confirm()) {
	    return null;
	}
	String fname = pathField.getText().trim();
	if (fname.length() == 0) {
	    return null;
	}
	int last = fname.length() - 1;
	if (last > 0 && fname.charAt(last) == File.separatorChar) {
	    return fname.substring(0, last);
	}
	return fname;
    }

    public String request(boolean showChooser) {
	if (showChooser) {
	    chooser.show();
	}
	if (!confirm()) {
	    return null;
	}
	String fname = pathField.getText().trim();
	if (fname.length() == 0) {
	    return null;
	}
	int last = fname.length() - 1;
	if (last > 0 && fname.charAt(last) == File.separatorChar) {
	    return fname.substring(0, last);
	}
	return fname;
    }

    public boolean handleEvent(Event ev) {
	if (ev.target == dirs) {
	    if (ev.id != Event.LIST_SELECT && ev.id != Event.LIST_DESELECT) {
		return true;
	    }
	    String name = dirs.getSelectedItem();
	    int selectedIndex = dirs.getSelectedIndex();
	    if (ev.id == Event.LIST_DESELECT ||
		selectedIndex == lastDirSelected ||
		name.equals(PARENT)) {
		int sep = currentDir.length() - 2;
		int drive = (SystemManager.onWin32 || SystemManager.onWin32) ?
			    currentDir.indexOf(":") : -1;
		if (sep > drive &&
		    (sep = currentDir.lastIndexOf(File.separator, sep)) >= 0) {
		    select(currentDir.substring(0, sep + 1));
		} else {
		    select(null);	// it is a top directory
		}
	    } else if (name.equals(CURRENT)) {
		select(null);	// redisplay current directory
	    } else {
		select(currentDir + name);
	    }
	    return true;
	} else if (ev.target == files) {
	    if (ev.id != Event.LIST_SELECT && ev.id != Event.LIST_DESELECT) {
		return true;
	    }
	    int selectedIndex = files.getSelectedIndex();
	    if (ev.id == Event.LIST_SELECT && selectedIndex != lastFileSelected) {
		select(currentDir + ((List)ev.target).getSelectedItem());
	    } else {
		select(currentDir + startFile);
	    }
	    return true;
	}
	return super.handleEvent(ev);
    }

    public boolean action(Event ev, Object arg) {
	if (ev.target == postfixField || ev.target == postfixButton) {
	    select(null);
	    return true;
	} else if (ev.target == pathField) {
	    String name = pathField.getText().trim();
	    int tail = name.length() - 1;
	    if (tail < 0) {
		select(null);
	    } else if (chooser.isVisible()) {
		TeikadeFile file = TeikadeFile.create(getPath());
		if (file.isDirectory()) {
		    currentFile = "";
		    setDirectory(getPath());
		} else {
		    setDirectory(file.getParent());
		    setFile(file.getName());
		}
	    } else {
		chosenFile = name;
		actionLeftButton();
	    }
	    return true;
	} else if (ev.target == chooseButton) {
	    if (chooser.isVisible()) {
		chooser.hide();
	    } else {
		chooser.show();
		TeikadeFile file = TeikadeFile.create(getPath());
		if (file.isDirectory()) {
		    currentFile = "";
		    setDirectory(getPath());
		} else {
		    setDirectory(file.getParent());
		    setFile(file.getName());
		}
	    }
	    validate();
	    pack();
	    layout();
	    return true;
	}
	return super.action(ev, arg);
    }

    private void select(String name) {
	lastDirSelected  = -1;
	lastFileSelected = -1;
	if (name == null || name.length() == 0) {
	    name = currentDir;
	}
	TeikadeFile f = TeikadeFile.create(name);
	boolean isDirectory = f.isDirectory();
	if (isDirectory) {
	    if (!name.endsWith(File.separator)) name += File.separator;
	    currentDir = name;
	    if (chooser.isVisible()) {
		if (dirs != null) {
		    dirs.clear();
		    dirs.addItem(CURRENT);
		    dirs.addItem(PARENT);
		}
		files.clear();
		String postfixString = postfixField.getText(); 
		int wild = postfixString.lastIndexOf('*');
		if (wild >= 0) postfixString = postfixString.substring(wild + 1);
		String list[] = StringSorter.sort(f.list());
		if (list != null)  {
		    for (int i = 0; i < list.length; i++) {
			String nm = list[i];
			f = TeikadeFile.create(currentDir + nm);
			if (f.isDirectory()) {
			    if (dirs != null) dirs.addItem(nm + File.separator);
			} else if (f.isFile() && nm.endsWith(postfixString)) {
			    files.addItem(nm);
			}
		    }
		}
		if (SystemManager.onSolaris)  {
		    Dimension d = files.size();
		    files.resize(d.width, d.height + 1);
		    files.resize(d.width, d.height);
		    if (dirs != null) {
			d = dirs.size();
			dirs.resize(d.width, d.height + 1);
			dirs.resize(d.width, d.height);
		    }
		}
	    }
	} else {
	    String nm = f.getName();
	    if (nm.length() > 0) {
		currentFile = f.getName();
	    }
	    if (chooser.isVisible()) {
		for (int i = files.countItems(); --i >= 0; ) {
		    if (files.getItem(i).equals(currentFile)) {
			if (!files.isSelected(i)) {
			    if (i + 3 < files.countItems()) {
				files.makeVisible(i + 3);
			    } else {
				files.makeVisible(i);
			    }
			    files.select(i);
			    lastFileSelected = i;
			}
			break;
		    }
		}
	    }
	}
	if (dirs == null) {
	    pathField.setText(currentFile);
	} else {
	    pathField.setText(currentDir + currentFile);
	}
    }

    public static
    String request(Frame frame, String title, String dir, String file) {
	TeikadeFileDialog dialog =  new TeikadeFileDialog(frame, title);
	dialog.setDirectory(dir);
	dialog.setFile(file);
	String path = dialog.request();
	dialog.dispose();
	return path;
    }

    public static
    String requestName(Frame frame, String title, String directory, String file) {

	TeikadeFileDialog dialog = new TeikadeFileDialog(frame, title, false);
	int postfixField = file.lastIndexOf('.');
	if (postfixField >= 0) {
	    dialog.setPostFix(file.substring(postfixField + 1));
	}
	dialog.setDirectory(directory);
	dialog.setFile(file);
	String name = dialog.request();
	dialog.dispose();
	return name;
    }
}
