/*
 * Copyright (c) 1996-1998 PFU Limited.
 *	author Kazuhisa Shirakami
 */
package teikade.gui.editor;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Stack;
import java.awt.*;
import teikade.gui.*;
import teikade.lang.CodeReader;
import teikade.lang.ForwardExpression;
import teikade.lang.BackwardExpression;
import teikade.lang.DescriptionException;
import teikade.util.TextUtil;

public class EditData {
    private EditDataViewer dataViewer;
    private EditDataObserver observer;
    private StatusViewer statusViewer;

    private String text;		// text data
    private int eot;			// end of text
    private int bol;			// beginning of line
    private int tol;			// text top of line
    private int eol;			// end of line
    private int current;		// caret position
    private int column;			// caret width
    private int targetColumn;		// column for previous/next line
    private int anchorMark;		// anchor mark
    private int lastAnchor;		// last anchor mark
    private int unnamedMark;		// unnamed mark
    private int lastKill;		// last kill point
    private UndoBuffer undoBuffer;
    private String statusMessage;
    private Object status;
    Object lastStatus;
    private String lastSearchString;
    String searchString = "";
    private Stack searchStack;

    private static String LF = System.getProperty("line.separator");
    static {
	if (LF == null) LF = "\n";
    }

    public EditData(EditDataViewer viewer) {
	dataViewer = viewer;
	undoBuffer = new UndoBuffer(this);
	reset();
    }

    public void reset() {
	undoBuffer.clear();
	targetColumn = -1;
	unnamedMark = -1;
	lastKill = -1;
	anchorMark = dataViewer.getSelectionStart();
	lastAnchor = anchorMark;
	setStatus(null, null);
	update();
    }

    public void setObserver(EditDataObserver observer) {
	this.observer = new EditDataDelegater(observer);
    }

    public boolean notifyObserver(int command) {
	if (observer != null) return observer.updateEditData(this, command);
	return false;
    }

    public boolean notifyObserverAsynchronous(int command) {
	if (observer != null) return new EditDataDelegater(observer).updateEditData(this, command);
	return false;
    }

    public void setStatusViewer(StatusViewer viewer) {
	statusViewer = viewer;
	undoBuffer.resetPoint();
    }

    public void setStatus(String message) {
	setStatus(message, null);
    }

    public void setStatus(String message, Object object) {
	statusMessage = message;
	status = object;
	if (statusViewer != null) {
	    statusViewer.viewStatus(message, object != null);
	}
    }

    public void pushStatus() {
	lastStatus = status;
	setStatus(null, null);
    }

    public void update() {
	text = dataViewer.getText();
	if (text != null) {
	    eot = text.length();
	} else {
	    text = "";
	    eot = 0;
	}
	update(dataViewer.getSelectionStart());
	lastAnchor = anchorMark;
	anchorMark = unnamedMark;
    }

    void update(int select) {
	if (select > eot) select = eot;
	current = select;
	bol = beginningOfLine(current);
	tol = endOfSpaces(bol);
	eol = endOfLine(current);
	column = TextUtil.getInstance().column(text.substring(bol, current),
					       EditProperties.hardTab);
    }

    static boolean isSpaceChar(char c) {
	return c == ' ' || c == '\t';
    }

    static boolean isEndOfLineChar(char c) {
	return c == '\n' || c == '\r';
    }

    public boolean onBeginningOfLine() {
	return current == bol;
    }

    public boolean onEndOfLine() {
	return current == eol;
    }

    public boolean onBeginningOfText() {
	return current == 0;
    }

    public boolean onEndOfText() {
	return current == eot;
    }

    int beginningOfLine(int pos) {
	while (pos > 0 && !isEndOfLineChar(text.charAt(pos - 1))) {
	    pos--;
	}
	return pos;
    }

    int beginningOfSpaces(int pos) {
	while (pos > 0 && isSpaceChar(text.charAt(pos - 1))) {
	    pos--;
	}
	return pos;
    }

    int endOfLine(int pos) {
	while (pos < eot && !isEndOfLineChar(text.charAt(pos))) {
	    pos++;
	}
	return pos;
    }

    int endOfSpaces(int pos) {
	while (pos < eot && isSpaceChar(text.charAt(pos))) {
	    pos++;
	}
	return pos;
    }

    int backwardChar(int pos) {
	if (pos > 0) {
	    pos--;
	    if (pos > 0
	    &&  text.charAt(pos - 1) == '\r' &&  text.charAt(pos) == '\n') {
		pos--;
	    }
	}
	return pos;
    }

    int forwardChar(int pos) {
	if (pos < eot) {
	    pos++;
	    if (pos < eot
	    &&  text.charAt(pos - 1) == '\r' &&  text.charAt(pos) == '\n') {
		pos++;
	    }
	}
	return pos;
    }

    int spacesAtBOL() {
	return spacesAtBOL(bol, tol);
    }

    int spacesAtBOL(int pos, int end) {
	int num = 0;
	TextUtil textUtil = TextUtil.getInstance();
	while (pos < end) {
	    char c = text.charAt(pos);
	    if (c == '\t') {
		num += EditProperties.hardTab;
		num -= num % EditProperties.hardTab;
	    } else if (textUtil.isDoubleWidth(c)) {
		num += 2;
	    } else {
		num++;
	    }
	    pos++;
	}
	return num;
    }

    static String spacesForBOL(int column) {
	StringBuffer spaces = new StringBuffer();
	while (column >= EditProperties.hardTab) {
	    spaces.append('\t');
	    column -= EditProperties.hardTab;
	}
	while (--column >= 0) {
	    spaces.append(' ');
	}
	return spaces.toString();
    }

    public void select(int pos) {
	select(pos, pos);
    }

    public void select(int start, int end) {
	dataViewer.select(start, end);
	update(start);
	undoBuffer.resetPoint();
	targetColumn = -1;
	lastKill = -1;
    }

    public void setText(String text) {
	dataViewer.setText(text);
	undoBuffer.clear();
	setStatus(null);
	update();
    }

    public void insertText(String text) {
	insertText(text, current);
    }

    public void insertText(String text, int pos) {
	int start = dataViewer.getSelectionStart();
	int end   = dataViewer.getSelectionEnd();
	if (start != end) {
	    undoBuffer.push(pos, getText(start, end), text);
	    dataViewer.replaceText("", start, end);
	} else {
	    undoBuffer.push(pos, "", text);
	}
	dataViewer.insertText(text, pos);
	int saveAnchor = anchorMark;
	update();
	anchorMark =  saveAnchor;
	select(pos + text.length());
	if (unnamedMark > pos) unnamedMark += text.length();
	targetColumn = -1;
	lastKill = -1;
    }

    public void replaceText(String text, int start, int end) {
	undoBuffer.push(start, getText(start, end), text);
	dataViewer.replaceText(text, start, end);
	update();
	select(start + text.length());
	if (unnamedMark > end) {
	    unnamedMark = unnamedMark - (end - start) + text.length();
	} else if (unnamedMark > start) {
	    unnamedMark = start;
	}
	targetColumn = -1;
	lastKill = -1;
    }

    public void deleteText(int start, int end) {
	if (start < end) {
	    replaceText("", start, end);
	} else if (start > end) {
	    replaceText("", end, start);
	}
    }

    public String getText() {
	return dataViewer.getText();
    }

    public String getText(int start, int end) {
	if (start < end) {
	    return text.substring(start, end);
	} else if (start > end) {
	    return text.substring(end, start);
	} else {
	    return "";
	}
    }

    public String getSelectedText() {
	String text = dataViewer.getSelectedText();
	if (text.length() > 0) {
	    return text;
	} else {
	    return null;
	}
    }

    public boolean cutSelectedText() {
	String selectedText = getSelectedText();
	if (selectedText != null) {
	    ClipBoard.setString(selectedText);
	    deleteSelectedText();
	    return true;
	} else {
	    return false;
	}
    }

    public void deleteSelectedText() {
	deleteText(dataViewer.getSelectionStart(), dataViewer.getSelectionEnd());
    }

    public void requestFocus() {
	dataViewer.requestFocus();
    }

    public void setEditable(boolean editable) {
	dataViewer.setEditable(editable);
    }

    public boolean isEditable() {
	return dataViewer.isEditable();
    }

    public int getSelectionStart() {
	return dataViewer.getSelectionStart();
    }

    public int getSelectionEnd() {
	return dataViewer.getSelectionEnd();
    }

    public int getColumn() {
	return column;
    }

    public int getTargetColumn() {
	return (targetColumn < 0) ? column : targetColumn;
    }

    public void setTargetColumn(int column) {
	targetColumn = column;
    }

    public void undo() {
	if (undoBuffer.undo()) {
	    setStatus("Undo!", null);	// null for no reset undo point
	} else {
	    setStatus("No further undo information");
	}
    }

    public void packUndo() {
	undoBuffer.pack();
    }

    public EditData setMark() {
	unnamedMark = current;
	undoBuffer.resetPoint();
	setStatus("Mark set");
	return this;
    }

    public EditData swapWithMark() {
	if (unnamedMark < 0) {
	    setStatus("The mark is not set now");
	} else {
	    int target = unnamedMark;
	    unnamedMark = current;
	    current = target;
	}
	return this;
    }

    public EditData anchorAtMark() {
	anchorMark = unnamedMark;
	return this;
    }

    public EditData anchorAtSelection() {
	if (lastAnchor == current) {
	    anchorMark = current;
	    update(getSelectionEnd());
	} else if (lastAnchor == getSelectionEnd()) {
	    anchorMark = lastAnchor;
	} else {
	    anchorMark = current;
	}
	return this;
    }

    public EditData anchor() {
	anchorMark = current;
	return this;
    }

    public EditData move() {
	select(current);
	return this;
    }

    public EditData moveToAnchor() {
	select(anchorMark);
	return this;
    }

    public EditData moveToColumn(int column) {
	gotoColumn(column);
	move();
	targetColumn = column;
	return this;
    }

    public EditData select() {
	if (anchorMark < current) {
	    select(anchorMark, current);
	} else {
	    select(current, anchorMark);
	}
	return this;
    }

    public EditData copy() {
	String selectedText = getSelectedText();
	if (EditProperties.selectedHandler == 2 ) {
	    if (selectedText != null) {
		ClipBoard.setString(selectedText);
	    } else if (anchorMark < 0) {
		return null;
	    } else {
		ClipBoard.setString(getText(anchorMark, current));
	    }
	} else {
	    if (selectedText != null) {
		ClipBoard.setString(selectedText);
	    } else {
		return null;
	    }
	}
	select(current);
	return this;
    }

    public EditData cut() {
	lastKill = -1;
	return kill();
    }

    public EditData kill() {
	String selectedText = getSelectedText();
	if (selectedText != null) {
	    anchorMark = dataViewer.getSelectionStart();
	    current = dataViewer.getSelectionEnd();
	}
	if (anchorMark < 0) {
	    return null;
	} else {
	    if (lastKill == anchorMark || lastKill == current) {
		ClipBoard.setString(ClipBoard.getString() + getText(anchorMark, current));
	    } else {
		ClipBoard.setString(getText(anchorMark, current));
	    }
	    deleteText(anchorMark, current);
	    lastKill = dataViewer.getSelectionStart();
	}
	return this;
    }

    public EditData appendKill() {
	String selectedText = getSelectedText();
	if (selectedText != null) {
	    anchorMark = dataViewer.getSelectionStart();
	    current = dataViewer.getSelectionEnd();
	}
	if (anchorMark < 0) {
	    return null;
	} else {
	    ClipBoard.setString(ClipBoard.getString() + getText(anchorMark, current));
	    deleteText(anchorMark, current);
	    lastKill = dataViewer.getSelectionStart();
	}
	return this;
    }

    public EditData delete() {
	deleteText(anchorMark, current);
	return this;
    }

    public EditData paste() {
	String str = ClipBoard.getString();
	if (str.length() > 0) {
	    update();
	    pushStatus();
	    setMark();
	    insertText(str, current);
	}
	return this;
    }

    public EditData insert(String text) {
	insertText(text, current);
	return this;
    }

    public EditData forwardChar() {
	if (onEndOfLine()) {
	    update(forwardChar(current));
	} else {
	    current = forwardChar(current);
	}
	return this;
    }

    public EditData backwardChar() {
	if (onBeginningOfLine()) {
	    update(backwardChar(current));
	} else {
	    current = backwardChar(current);
	}
	return this;
    }

    public EditData beginningOfSpaces() {
	current = beginningOfSpaces(current);
	return this;
    }

    public EditData beginningOfLetterOrDigit() {
	while (current > 0 && Character.isJavaLetterOrDigit(text.charAt(current - 1))) {
	    current--;
	}
	return this;
    }

    public EditData beginningOfUnLetterOrDigit() {
	while (current > 0 && !Character.isJavaLetterOrDigit(text.charAt(current - 1))) {
	    current--;
	}
	return this;
    }

    public EditData beginningOfLine() {
	current = bol;
	return this;
    }

    public EditData beginningOfText() {
	current = 0;
	return this;
    }

    public EditData endOfSpaces() {
	current = endOfSpaces(current);
	return this;
    }

    public EditData endOfLetterOrDigit() {
	while (current < eot && Character.isJavaLetterOrDigit(text.charAt(current))) {
	    current++;
	}
	return this;
    }

    public EditData endOfUnLetterOrDigit() {
	while (current < eot && !Character.isJavaLetterOrDigit(text.charAt(current))) {
	    current++;
	}
	return this;
    }

    public EditData endOfLine() {
	current = eol;
	return this;
    }

    public EditData endOfText() {
	current = eot;
	return this;
    }

    public EditData previousLine() {
	return previousLine(1);
    }

    public EditData previousLine(int n) {
	if (n > 0) {
	    int targetColumn = getTargetColumn();
	    while (--n >= 0 && bol > 0) {
		current = backwardChar(beginningOfLine(current));
	    }
	    moveToColumn(targetColumn);
	} else if (n < 0) {
	    nextLine(-n);
	}
	return this;
    }

    public EditData nextLine() {
	return nextLine(1);
    }

    public EditData nextLine(int n) {
	if (n > 0) {
	    int targetColumn = getTargetColumn();
	    while (--n >= 0 && eol < eot) {
		current = forwardChar(endOfLine(current));
	    }
	    moveToColumn(targetColumn);
	} else if (n < 0) {
	    previousLine(-n);
	}
	return this;
    }

    public EditData tab(boolean backword) {
	if (!backword) {	// forward TAB
	    if (current < tol) {
		select(tol);
	    } else if (current == tol) {
		int column = spacesAtBOL();
		column += EditProperties.softTab - (column % EditProperties.softTab);
		replaceText(spacesForBOL(column), bol, tol);
	    } else {
		insertText("\t", current);
	    }
	} else {	// backword TAB
	    if (current > tol) {
		select(tol);
	    } else if (current > bol) {
		int column = spacesAtBOL() - 1;
		column -= column % EditProperties.softTab;
		replaceText(spacesForBOL(column), bol, tol);
	    } else {
		// nothing to do at beginning of line
	    }
	}
	return this;
    }

    public EditData forwardExpression() {
	if (text == null) return null;
	ForwardExpression exp = new ForwardExpression(text, current);
	try {
	    exp.nextExpression();
	} catch (DescriptionException e) {
	    setStatus(e.getMessage());
	    return null;
	}
	current = exp.endPoint();
	return this;
    }

    public EditData backardExpression() {
	if (text == null) return null;
	BackwardExpression exp = new BackwardExpression(text, current);
	try {
	    exp.nextExpression();
	} catch (DescriptionException e) {
	    setStatus(e.getMessage());
	    return null;
	}
	current = exp.endPoint();
	return this;
    }

    public EditData forwardPair() {
	if (text == null) return null;
	ForwardExpression exp = new ForwardExpression(text, current);
	try {
	    exp.nextPair();
	} catch (DescriptionException e) {
	    setStatus(e.getMessage());
	    return null;
	}
	current = exp.endPoint();
	return this;
    }

    public EditData backwardPair() {
	if (text == null) return null;
	BackwardExpression exp = new BackwardExpression(text, current);
	try {
	    exp.nextPair();
	} catch (DescriptionException e) {
	    setStatus(e.getMessage());
	    return null;
	}
	current = exp.endPoint();
	return this;
    }

    public EditData autoIndent() {
	if (bol == 0) return tab(false);
	CodeReader reader = new CodeReader(text, 99, false);
	int nest, lastNest = 0;
	try {
	    while ((nest = reader.nextDescription()) > 0) {
		if (eol < reader.endPoint() || eot == reader.endPoint()) break;
		lastNest = nest;
	    }
	}
	catch (teikade.lang.DescriptionException e) {
	    nest = lastNest;
	}
	if (nest > lastNest) {
	    if (bol <= reader.keyPoint()) nest--;
	} else if (nest < lastNest) {
	    if (eol <= reader.keyPoint()) nest++;
	}
	if (startsWithCaseOrDefault()) {
	    nest--;
	}
	if (nest > 0 || reader.startPoint() == 0 || bol <= reader.keyPoint()) {
	    nest += spacesAtBOL(0, endOfSpaces(0)) / EditProperties.softTab;
	}
	String tabs = spacesForBOL(nest * EditProperties.softTab);
	boolean needReplace = !tabs.equals(getText(bol, tol));
	if (current < tol) {
	    if (needReplace) replaceText(tabs, bol, tol);
	    select(tol);
	} else {
	    int pos = current - (tol - bol) + tabs.length();
	    if (needReplace) replaceText(tabs, bol, tol);
	    select(pos);
	}
	return this;
    }

    public EditData setSearchString(String text) {
	lastSearchString = searchString;
	searchString = text;
	return this;
    }

    public EditData setLastSearchString() {
	pushSearchString();
	searchString = lastSearchString;
	return this;
    }

    public EditData clearSearchString() {
	lastSearchString = searchString;
	searchString = "";
	searchStack = new Stack();
	return this;
    }

    public EditData quitSearch() {
	if (!searchStack.empty()) {
	    current = ((SearchData)searchStack.firstElement()).point;
	    searchStack = null;
	}
	return this;
    }

    public EditData addSearchString(String text) {
	pushSearchString();
	lastSearchString = searchString;
	searchString += text;
	return this;
    }

    public EditData pushSearchString() {
	searchStack.push(new SearchData(searchString, current));
	return this;
    }

    public EditData backspaceSearchString() {
	if (!searchStack.empty()) {
	    lastSearchString = searchString;
	    SearchData data = (SearchData)searchStack.pop();
	    searchString = data.string;
	    current = data.point;
	}
	return this;
    }

    public String getSearchString() {
	return searchString;
    }

    public EditData forwardSearch() {
	int pos = current;
	if (searchString.length() == 0) {
	    searchString = lastSearchString;
	} else if (searchString.equals(lastSearchString)) {
	    pos++;
	}
	pos = text.indexOf(searchString, pos);
	if (pos < 0) return null;
	current = pos;
	anchorMark = pos + searchString.length();
	return this;
    }

    public EditData backwardSearch() {
	int pos = current;
	if (searchString.length() == 0) {
	    searchString = lastSearchString;
	} else if (searchString.equals(lastSearchString)) {
	    pos--;
	}
	pos = text.lastIndexOf(searchString, pos);
	if (pos < 0) return null;
	current = pos;
	anchorMark = pos + searchString.length();
	return this;
    }

    public boolean startsWithCaseOrDefault() {
	return text.startsWith("case ", tol)
	    || text.startsWith("case\t", tol)
	    || text.startsWith("default ", tol)
	    || text.startsWith("default\t", tol)
	    || text.startsWith("default:", tol);
    }

    public EditData returnAndIndent() {
	String spaces = spacesForBOL(spacesAtBOL());
	insertText(LF);
	if (spaces.length() > 0) {
	    insertText(spaces);
	}
	return this;
    }

    public EditData backSpace() {
	if (current == 0) return this;
	if (bol < current && current <= tol) {
	    int column = spacesAtBOL();
	    if ((column % EditProperties.softTab) == 0) {
		replaceText(spacesForBOL(column - EditProperties.softTab), bol, current);
		return this;
	    }
	}
	return anchor().backwardChar().delete();
    }

    EditData gotoColumn(int target) {
	update(current);
	TextUtil textUtil = TextUtil.getInstance();
	if (textUtil.column(text.substring(bol, eol),
			    EditProperties.hardTab) <= target) {
	    current = eol;
	} else {
	    int pos = bol;
	    int col = 0;
	    while (pos < eol) {
		char c = text.charAt(pos);
		if (c == '\t') {
		    col += EditProperties.hardTab;
		    col -= col % EditProperties.hardTab;
		} else if (textUtil.isDoubleWidth(c)) {
		    col += 2;
		} else {
		    col++;
		}
		if (col > target) break;
		pos++;
	    }
	    current = pos;
	}
	return this;
    }

    public EditData previousPage() {
	previousLine(dataViewer.displayLines());
	return this;
    }

    public EditData nextPage() {
	nextLine(dataViewer.displayLines());
	return this;
    }
}

class EditDataDelegater implements EditDataObserver, Runnable {
    private EditDataObserver observer;
    private Vector data;
    private Thread delegater;

    EditDataDelegater(EditDataObserver observer) {
	this.observer = observer;
	this.data = new Vector();
	this.delegater = null;
    }

    public synchronized boolean updateEditData(EditData target, int command) {
	data.addElement(new DelegateData(target, command));
	if (delegater == null) (delegater = new Thread(this,this.getClass().getName())).start();
	return true;
    }

    private synchronized DelegateData getCommand() {
	if (data.size() == 0) return null;
	DelegateData d = (DelegateData)data.firstElement();
	data.removeElementAt(0);
	return d;
    }

    public void run() {
	DelegateData data;
	while ((data = getCommand()) != null) {
	    observer.updateEditData(data.target, data.command);
	}
	delegater = null;
    }
}

class DelegateData {
    EditData target;
    int command;

    DelegateData(EditData t, int c) {
	target = t;
	command = c;
    }
}

class SearchData {
    String string;
    int point;

    SearchData(String s, int p) {
	string = s;
	point = p;
    }
}
