/*
 * Copyright (c) 1995-1998 PFU Limited.
 *	author Kazuhisa Shirakami
 */
package teikade.util;

import java.io.*;
import teikade.sys.SystemManager;
import teikade.sys.TeikadeProperties;
import teikade.util.CodeConverter;

public class CommandExternal implements Command, Runnable {
    private static final boolean DEBUG =
	"true".equals(System.getProperty("debug.execute"));
    private static final String DEBUG_HEADER = "debug.execute: ";
    private String command[];
    private Process process;
    private boolean doOutput = true, doError = true, asynchronous = true;
    private Thread executeThread , errorThread;
    private CommandExternalListener listener;

    private static String shell;
    private static String shellOption;
    public static boolean useShell = false;
    public static boolean useTeikadeShell = false;
    static {
	if (SystemManager.onWin32) {
	    shell = SystemManager.TEIKADEBIN + "dash.exe";
	    if (new File(shell).exists()) {
		useTeikadeShell = true;
	    } else if (SystemManager.onWinNT) {
		shell = "c:\\windows\\cmd.exe";
	    } else {
		shell = "c:\\windows\\command.com";
	    }
	    shellOption = "/C";
	    useShell = true;
	} else if (!SystemManager.onJDK1_1) {
	    shell = "/bin/sh";
	    if (new File(shell).exists()) {
		shellOption = "-c";
		useShell = true;
	    }
	}
    }

    public static String getJDKCommandName(String name) {
	String suffix = System.getProperty("command.suffix");
	if (suffix == null) {
	    if (SystemManager.onWin32 || SystemManager.onOS2) {
		suffix = ".exe";
	    } else {
		suffix = "";
	    }
	}
	String home = System.getProperty("java.home");
	String dir  = File.separator + "bin"
		    + File.separator + name
		    + suffix;
	String command = home + dir;
	if (new File(command).exists()) {
	    return command;
	}
	command = home + File.separator + ".." + dir;
	if (new File(command).exists()) {
	    return command;
	}
	return null;
    }

    protected CommandExternal() {
    }

    public CommandExternal(String cmd, CommandExternalListener listener) {
	setCommand(cmd);
	setCommandExternalListener(listener);
    }

    public CommandExternal(String[] cmd, CommandExternalListener listener) {
	setCommand(cmd);
	setCommandExternalListener(listener);
    }

    public void setCommand(String[] cmd) {
	if (cmd == null || cmd.length == 0
	 || cmd[0] == null || cmd[0].length() == 0) {
	    throw new IllegalArgumentException("no command");
	};
	if (useShell) {
	    StringBuffer buf = new StringBuffer();
	    buf.append(cmd[0]);
	    for (int i = 1; i < cmd.length; i++) {
		buf.append(' ').append(cmd[i]);
	    }
	    setCommandUsingShell(buf.toString());
	} else {
	    command = cmd;
	}
    }

    public void setCommand(String cmd) {
	if (cmd == null || cmd.length() == 0) {
	    throw new IllegalArgumentException("no command");
	}
	if (useShell) {
	    setCommandUsingShell(cmd);
	} else {
	    command = new String[1];
	    command[0] = cmd;
	}
    }

    protected void setCommandUsingShell(String cmd) {
	command = new String[3];
	command[0] = shell;
	command[1] = shellOption;
	if (SystemManager.onWin32) {
	    if (useTeikadeShell) {
		int escape = cmd.length();
		escape = cmd.length();
		while ((escape = cmd.lastIndexOf('"', escape - 1)) >= 0) {
		    cmd = cmd.substring(0, escape) + "\\" + cmd.substring(escape);
		}
		setDoError(false);
	    }
	    command[2] = cmd;
	} else {
	    command[2] = cmd + " 2>&1";
	    setDoError(false);
	}
    }

    protected void setCommandExternalListener(CommandExternalListener listener) {
	if (listener == null) {
	    throw new IllegalArgumentException("listener is null");
	}
	this.listener = listener;
    }

    public void setDoOutput(boolean doOutput) {
	this.doOutput = doOutput;
    }

    public void setDoError(boolean doError) {
	this.doError = doError;
    }

    public void setAsynchronous(boolean async) {
	asynchronous = async;
    }

    public void execute() {
	if (process != null) {
	    throw new IllegalThreadStateException("command is running");
	}
	if (asynchronous) {
	    (executeThread = new Thread(this,this.getClass().getName()+"-Execute")).start();
	} else {
	    doExecute();
	}
    }

    public void run() {
	Thread currentThread = Thread.currentThread();
	if (currentThread == executeThread) {
	    doExecute();
	} else if (currentThread == errorThread) {
	    doError(process.getErrorStream());
	}
    }

    protected void doExecute() {
	if (listener != null) {
	    listener.executing(this);
	}
	try {
	    if (DEBUG) {
		System.err.print(DEBUG_HEADER);
		for (int i = 0; i < command.length; i++) {
		    System.err.print(" \"");
		    System.err.print(command[i]);
		    System.err.print("\"");
		}
		System.err.print(" : ");
	    }
	    process = Runtime.getRuntime().exec(command);
	    if (DEBUG) {
		System.err.println(process);
	    }
	} catch (IOException e) {
	    if (DEBUG) {
		System.err.println("Runtime.exec() caught IOException");
	    }
	    if (listener != null) {
		listener.aborted(this, e);
	    }
	    return;
	}
	if (doError && listener != null) {
	    (errorThread = new Thread(this,this.getClass().getName()+"-Error")).start();
	}
	if (doOutput && listener != null) {
	    doOutput(process.getInputStream());
	}
	doWait();
    }

    protected void doOutput(InputStream outputStream) {
	if (DEBUG) {
	    System.err.print(DEBUG_HEADER);
	    System.err.println("do output");
	}
	DataInputStream stream = new DataInputStream(outputStream);
	try {
	    String line;
	    while ((line = CodeConverter.readLine(stream)) != null) {
		if (DEBUG) {
		    System.err.print(DEBUG_HEADER);
		    System.err.print("output: ");
		    System.err.println(line);
		}
		listener.output(this, line);
	    }
	} catch (IOException e) {
	    if (DEBUG) {
		System.err.print(DEBUG_HEADER);
		System.err.println("doOuput() caught IOException");
	    }
	}
    }

    protected void doError(InputStream errorStream) {
	if (DEBUG) {
	    System.err.print(DEBUG_HEADER);
	    System.err.println("do error");
	}
	try {
	    String line;
	    DataInputStream stream = new DataInputStream(errorStream);
	    while (Thread.currentThread() == errorThread
	       &&  (line = CodeConverter.readLine(stream)) != null) {
		if (!line.equals("err = 1!")) {
		if (DEBUG) {
		    System.err.print(DEBUG_HEADER);
		    System.err.print("error: ");
		    System.err.println(line);
		}
		    listener.error(this, line);
		}
	    }
	} catch (IOException e) {
	    if (DEBUG) {
		System.err.print(DEBUG_HEADER);
		System.err.println("doError() caught IOException");
	    }
	} finally {
	    synchronized (this) {
		errorThread = null;
		notifyAll();
	    }
	}
    }

    protected void doWait() {
	try {
	    process.waitFor();
	} catch (InterruptedException e) {
	    if (DEBUG) {
		System.err.print(DEBUG_HEADER);
		System.err.println("Process.waitFor() caught InterruptedException");
	    }
	    if (listener != null) {
		listener.aborted(this, e);
	    }
	    process = null;
	    return;
	}
	while (errorThread != null) {
	    synchronized (this) {
		try {
		    wait();
		} catch (InterruptedException e) {
		    if (DEBUG) {
			System.err.print(DEBUG_HEADER);
			System.err.println("wait() caught InterruptedException");
		    }
		    if (listener != null) {
			listener.aborted(this, e);
		    }
		    process = null;
		    return;
		}
	    }
	}
	int exitValue;
	try {
	    exitValue = process.exitValue();
	    if (DEBUG) {
		System.err.print(DEBUG_HEADER);
		System.err.print("exit value = ");
		System.err.println(exitValue);
	    }
	    if (listener != null) {
		listener.finished(this, exitValue);
	    }
	} catch (IllegalThreadStateException e) {
	    if (DEBUG) {
		System.err.print(DEBUG_HEADER);
		System.err.println("Process.exitValue() caught IllegalThreadStateException");
	    }
	    listener.aborted(this, e);
	} finally {
	    process = null;
	}
    }

    public synchronized void abort() {
	if (process != null) {
	    if (DEBUG) {
		System.err.print(DEBUG_HEADER);
		System.err.print("aborting ");
		System.err.println(process);
	    }
	    process.destroy();
	    try {
		Thread.sleep(100);
	    } catch (InterruptedException e) {
	    }
	}
	if (errorThread != null) {
	    Thread thread = errorThread;
	    errorThread = null;
	    thread.interrupt();
	}
	if (executeThread != null) {
	    Thread thread = executeThread;
	    executeThread = null;
	    thread.interrupt();
	}
    }
}
